from learns2.constants import races
from learns2.parser import SC2ReplayParser
from collections import deque, defaultdict
from collections.abc import Iterator
from typing import List
import numpy


class EventIterator(Iterator):
    """Groups game events by game loop."""

    def __init__(self, events: List, num_frames: int):
        """
        :param events: raw sequence of game events
        :param num_frames: number of frames to return
        """
        self.events = deque(events)
        self.buffer = None
        self.frame = 0
        self.num_frames = num_frames

    def __next__(self):
        """Returns a list of all game events for the current frame."""
        if self.frame == self.num_frames:
            raise StopIteration
        buf = []
        while self.events and self.events[0]['_gameloop'] == self.frame:
            buf.append(self.events.popleft())
        self.frame += 1
        if not self.events and not buf:
            raise StopIteration
        return buf


def preprocess(replay, num_frames):
    parser = SC2ReplayParser(replay)
    players = parser.players()
    itr = EventIterator(parser.events(), num_frames)
    return players, list(itr)


# TODO numpy everything
class SC2ReplayFeaturizer(object):
    MAX_SELECTION_SIZE = 1000

    def feature_shape(self):
        return self.num_frames, 15 + self.num_camera_hotspots

    def __init__(self, replay, user_id: int, num_frames: int = 500, num_camera_hotspots: int = 5):
        self.players, self.frames = preprocess(replay, num_frames)
        self.num_frames = num_frames
        self.user_id = user_id
        self.num_camera_hotspots = num_camera_hotspots

    def feature(self):
        f1 = self.hotkey_feature()
        f2 = self.target_feature()
        f3 = self.races_feature()
        f4 = self.selection_feature()
        f5 = self.camera_hotspots_feature()
        f6 = self.scmd_feature()
        features = []
        for (i, j, k, x, y, z) in zip(f1, f2, f3, f4, f5, f6):
            features.append(i + j + k + x + y + z)

        arr = numpy.array(features)
        assert arr.shape == self.feature_shape(), 'Bad feature shape!'
        return arr

    def scmd_feature(self):
        features = []
        for frame in self.frames:
            feature = [0]
            for event in frame:
                if event['_event'] == 'NNet.Game.SCmdEvent' \
                        and event['_userid']['m_userId'] == self.user_id:
                    feature = [1]
            features.append(feature)
        return features

    def hotkey_feature(self):
        features = []
        for frame in self.frames:
            feature = [0] * 10
            for event in frame:
                if event['_event'] == 'NNet.Game.SControlGroupUpdateEvent' \
                        and event['_userid']['m_userId'] == self.user_id:
                    idx = event['m_controlGroupIndex']
                    feature[idx] = 1
            features.append(feature)
        return features

    def target_feature(self):
        features = []
        for frame in self.frames:
            feature = [0, 0]
            for event in frame:
                if event['_userid']['m_userId'] == self.user_id:
                    if event['_event'] == 'NNet.Game.SCmdUpdateTargetPointEvent':
                        feature[0] = 1
                    elif event['_event'] == 'NNet.Game.SCmdUpdateTargetUnitEvent':
                        feature[1] = 1
            features.append(feature)
        return features

    def races_feature(self):
        feature = [0]
        for p in self.players:
            if p['m_userId'] == self.user_id:
                race = races[p['m_race']]
                if race == 'Protoss':
                    feature[0] = 0.2
                elif race == 'Terran':
                    feature[0] = 0.4
                elif race == 'Zerg':
                    feature[0] = 0.6
        return [feature for _ in self.frames]

    def selection_feature(self):
        features = []
        for frame in self.frames:
            # This potentially overwrites earlier events in the rare case where there are multiple selection delta
            # events in a single gameloop. A quick scan of my replay folder found 0 occurrences of this, so I'm going to
            # ignore that case in the interest of simplicity.
            feature = [0]
            for event in frame:
                if event['_event'] == 'NNet.Game.SSelectionDeltaEvent' and event['_userid']['m_userId'] == self.user_id:
                    num_selected = len(event['m_delta']['m_addUnitTags'])
                    feature = [num_selected / self.MAX_SELECTION_SIZE]
            features.append(feature)
        return features

    def camera_hotspots_feature(self):
        # Count the number of times the player visited each (x, y) location
        locations = defaultdict(int)
        for frame in self.frames:
            for event in frame:
                if event['_event'] == 'NNet.Game.SCameraUpdateEvent' and event['_userid']['m_userId'] == self.user_id:
                    x, y = event['m_target']['x'], event['m_target']['y']
                    locations[(x, y)] += 1

        # Select the top n visited locations
        sorted_locations = {k: v for k, v in sorted(locations.items(), key=lambda item: item[1], reverse=True)}
        hotspots = [k for k in sorted_locations][:self.num_camera_hotspots]

        # Each feature is an array of length `self.num_camera_hotspots`
        # Each value in the array represents an individual camera hotspot
        # For each gameloop, set the features value to 1 if the user's camera matches the coordinates of the hotspot
        features = []
        for frame in self.frames:
            feature = [0] * self.num_camera_hotspots
            for event in frame:
                if event['_event'] == 'NNet.Game.SCameraUpdateEvent' and event['_userid']['m_userId'] == self.user_id:
                    x, y = event['m_target']['x'], event['m_target']['y']
                    for (idx, hotspot) in enumerate(hotspots):
                        if hotspot == (x, y):
                            feature[idx] = 1
            features.append(feature)

        return features



from s2protocol import versions
from mpyq import MPQArchive
from typing import List
import os


class SC2ReplayParser(object):
    def __init__(self, replay: str):
        self.archive = MPQArchive(replay)
        self.protocol = self.read_protocol(self.archive)
        self._initdata = None
        self._details = None
        self._events = None
        self._attributeevents = None
        self._trackerevents = None
        self._metadata = None

    @staticmethod
    def read_protocol(archive: MPQArchive):
        content = archive.header['user_data_header']['content']
        header = versions.latest().decode_replay_header(content)
        base_build = header['m_version']['m_baseBuild']
        try:
            return versions.build(base_build)
        # https://github.com/Blizzard/s2protocol/issues/99
        # If the protocol is not defined, the previous protocol often works anyway
        except ImportError as e:
            # print(f'Unable to import protocol{base_build}.py. Module does not exist')
            base_path = os.path.dirname(versions.__file__)
            protocols = set([p for p in os.listdir(base_path)])
            while base_build > 0:
                base_build -= 1
                candidate = f'protocol{base_build}.py'
                if candidate in protocols:
                    # print(f'Falling back to {candidate}')
                    return versions.build(base_build)
            raise e

    @staticmethod
    def read_initdata(protocol, archive: MPQArchive) -> dict:
        init_file = archive.read_file('replay.initData')
        return protocol.decode_replay_initdata(init_file)

    @staticmethod
    def read_details(protocol, archive: MPQArchive) -> dict:
        detail_file = archive.read_file('replay.details')
        return protocol.decode_replay_details(detail_file)

    @staticmethod
    def read_game_events(protocol, archive: MPQArchive) -> dict:
        game_event_file = archive.read_file('replay.game.events')
        return protocol.decode_replay_game_events(game_event_file)

    @staticmethod
    def read_attribute_events(protocol, archive: MPQArchive) -> dict:
        attrib_event_file = archive.read_file('replay.attributes.events')
        return protocol.decode_replay_attributes_events(attrib_event_file)

    @staticmethod
    def read_tracker_events(protocol, archive: MPQArchive) -> dict:
        tracker_event_file = archive.read_file('replay.tracker.events')
        return protocol.decode_replay_tracker_events(tracker_event_file)

    @staticmethod
    def read_metadata(protocol, archive: MPQArchive) -> dict:
        return archive.read_file('replay.gamemetadata.json')

    def initdata(self) -> dict:
        if self._initdata is None:
            self._initdata = self.read_initdata(self.protocol, self.archive)
        return self._initdata

    def details(self) -> dict:
        """https://github.com/Blizzard/s2protocol/blob/master/docs/flags/details.md"""
        if self._details is None:
            self._details = self.read_details(self.protocol, self.archive)
        return self._details

    def events(self):
        if self._events is None:
            self._events = list(self.read_game_events(self.protocol, self.archive))
        return self._events

    def attributeevents(self):
        if self._attributeevents is None:
            self._attributeevents = list(self.read_attribute_events(self.protocol, self.archive))
        return self._attributeevents

    def trackerevents(self):
        if self._trackerevents is None:
            self._trackerevents = list(self.read_tracker_events(self.protocol, self.archive))
        return self._trackerevents

    def metadata(self):
        if self._metadata is None:
            import json
            self._metadata = json.loads(self.read_metadata(self.protocol, self.archive))
        return self._metadata

    def players(self) -> List[dict]:
        slots = self.initdata()['m_syncLobbyState']['m_lobbyState']['m_slots']
        slot_id_to_user_id = {}
        for slot in slots:
            if slot['m_observe'] == 0:
                slot_id, user_id = slot['m_workingSetSlotId'], slot['m_userId']
                slot_id_to_user_id[slot_id] = user_id
        player_list = self.details()['m_playerList']
        player_objects = []
        for player in player_list:
            slot_id = player['m_workingSetSlotId']
            if slot_id in slot_id_to_user_id:
                userid = slot_id_to_user_id[slot_id]
                player['m_userId'] = userid
                player['m_userInitialData'] = self.initdata()['m_syncLobbyState']['m_userInitialData'][userid]
                player['m_localizedId'] = '{}/{}/{}'.format(player['m_toon']['m_region'], player['m_toon']['m_realm'],
                                                            player['m_toon']['m_id'])
                player_objects.append(player)
        return [decode_utf8(obj) for obj in player_objects]

    def small_details(self):
        details = self.details()
        description = self.initdata()['m_syncLobbyState']['m_gameDescription']
        return {
            'm_title': details['m_title'],
            'm_isBlizzardMap': details['m_isBlizzardMap'],
            'm_timeUTC': details['m_timeUTC'],
            'm_timeLocalOffset': details['m_timeLocalOffset'],
            'm_mapSizeX': description['m_mapSizeX'],
            'm_mapSizeY': description['m_mapSizeY']
        }

    def to_dict(self):
        payload = {
            'isLabeled': False,
            'labels': {},
            'details': self.small_details(),
            'players': self.players(),
            'metadata': self.metadata()
        }
        return decode_utf8(payload)


# Slightly modified from:
# https://stackoverflow.com/questions/57014259/json-dumps-on-dictionary-with-bytes-for-keys
def decode_utf8(d: dict) -> dict:
    result = {}
    for key, value in d.items():
        try:
            if isinstance(key, bytes):
                key = key.decode()
            if isinstance(value, bytes):
                value = value.decode()
            elif isinstance(value, dict):
                value = decode_utf8(value)
            elif isinstance(value, list):
                decoded = [decode_utf8(v) for v in value]
                value = decoded
        except:
            value = ''
        result.update({key: value})
    return result



races = {
    'Zerg': 'Zerg',
    '저그': 'Zerg',
    '异虫': 'Zerg',
    '蟲族': 'Zerg',
    '테란': 'Terran',
    '人類': 'Terran',
    '人类': 'Terran',
    'Terran': 'Terran',
    'Terraner': 'Terran',
    'Терраны': 'Terran',
    '프로토스': 'Protoss',
    '神族': 'Protoss',
    'Protosi': 'Protoss',
    '星灵': 'Protoss',
    'Протоссы': 'Protoss',
    'Protoss': 'Protoss'
}
